import * as zlib from 'zlib';
import { ISsnFileEntry } from '../interfaces/ISsnFileEntry';
import decryptFile from './decryption/decryptFile';
import ByteReader from './extractFileByteReader';

/** Extracts the given file from the given DataView array and returns it as an ArrayBuffer.
 * Will throw an error when end of final DataView is reached.
 */
export default async function extractFile(file: ISsnFileEntry, dvArray: DataView[]): Promise<ArrayBuffer> {
  //use ByteReader for reading a uint8 and seeking forward across DataView boundaries
  const byteReader = new ByteReader(dvArray, file.diskNumberStart, file.offset);

  if (byteReader.readByte() !== 0x50 || byteReader.readByte() !== 0x4B || byteReader.readByte() !== 0x03 || byteReader.readByte() !== 0x04) {
        throw new Error('Local file header had wrong magic');
  }
  byteReader.seek(22);
  const localFilenameSize = byteReader.readByte() + (byteReader.readByte() << 8);
  const localExtraSize = byteReader.readByte() + (byteReader.readByte() << 8);
  byteReader.seek(localFilenameSize + localExtraSize);

  let dvFinal = byteReader.extractDv(file.comprSize);

  //decrypt file if necessary
  if (file.decryptionKeys !== undefined) {
    dvFinal = decryptFile(dvFinal, file.comprSize, file.decryptionKeys);
  }

  //uncompress file
  const uncompressedBuffer: Buffer = await new Promise((resolve, reject) => {
    zlib.inflateRaw(dvFinal, (error, result) => {
      if (error !== null) {
        reject(error);
      }
      resolve(result);
    });
  }) as Buffer;
  return uncompressedBuffer.buffer as ArrayBuffer;
}
