import { TextDecoder } from 'util';
import * as xmlJs from 'xml-js';
import getUrlContents from '../cdn/getUrlContents';
import IManifest from '../interfaces/IManifest';
import { Product } from '../interfaces/ISettings';
import parsePatchmanifest from '../ssn/parsePatchmanifest';
import verifyPatchmanifest from '../ssn/verifyPatchmanifest';
import verifyProductName from '../ssn/verifyProductName';
import extractFile from './extractFile';
import readSsnFile from './readSsnFile';

const Decoder = new TextDecoder('utf-8');

export default async function getPatchmanifest(product: Product): Promise<IManifest> {
  //Verify function arguments
  if (!verifyProductName(product)) {
    throw new Error(`"${product}" is not a valid product.`);
  }

  //Download .patchmanifest file
  const ssnFile = await getUrlContents({ host: 'manifest.swtor.com', path: `/patch/${product}.patchmanifest` });

  //Parse .patchmanifest file
  const fileEntries = readSsnFile(ssnFile);

  //Verify .patchmanifest file
  if (fileEntries.length !== 1) {
    throw new Error(`Expected .patchmanifest to contain 1 file but it had "${fileEntries.length}" files.`);
  }

  const firstFile = fileEntries[0];
  if (firstFile.name !== 'manifest.xml') {
    throw new Error(`Expected .patchmanifest to contain a file called manifest.xml but it is called "${firstFile.name}".`);
  }

  //Extract manifest.xml file
  const patchmanifestFile = await extractFile(firstFile, [new DataView(ssnFile)]);

  //Convert ArrayBuffer to string
  const patchmanifestXml = Decoder.decode(patchmanifestFile);

  //convert XML to JSON-converted XML
  const patchManifestJson = xmlJs.xml2js(patchmanifestXml) as xmlJs.Element;

  //verify that XML file contains all required elements and attributes
  verifyPatchmanifest(patchManifestJson, product);

  //convert JSON-converted XML to an easier to read JSON
  const patchManifestSimple = parsePatchmanifest(patchManifestJson);

  return patchManifestSimple;
}
